/* $Id: cut_test1.c,v 1.12 2012/09/13 08:12:31 zlb Exp $ */

#include "phg.h"
#include <math.h>
#include <string.h>

#define dist2(v,vc) \
	((vc) == NULL \
		? v[0]*v[0] + v[1]*v[1] + v[2]*v[2] \
		: Pow(v[0]-((FLOAT *)vc)[0],2) + \
		  Pow(v[1]-((FLOAT *)vc)[1],2) + \
		  Pow(v[2]-((FLOAT *)vc)[2],2))
#define dist(v,vc) Sqrt(dist2(v, vc))

static FLOAT
getCutPosition(FLOAT *v0, FLOAT *v1, FLOAT *center, FLOAT r, FLOAT tol)
{
    FLOAT v[3];
#if 0
#define EPS	(FLOAT_EPSILON * 10.0)
    /* compute cuts using bi-section */
    BOOLEAN swap = FALSE;
    FLOAT down, up, mid, d;

    assert(r > tol);

    down = dist(v0, center);
    up = dist(v1, center);

    /* check wether v0 or v1 is on the surface */
    if (Fabs(down - r) <= EPS || Fabs(up - r) < EPS)
	return -1.0;

    /* make sure dist(v0,center) < dist(v1,center), swap them if necessary */
    if (down > up) {
	/* swap v0 and v1 */
	FLOAT *v = v0; v0 = v1; v1 = v;
	d = down; down = up; up = d;
	swap = TRUE;
    }

    /* no cut if both v0 and v1 are inside or outside the sphere */
    if (down >= r + EPS || up <= r - EPS)
	return -1;

    /* now v0 is inside and v1 is outside the sphere */
    down = 0.0; up = 1.0;
    while (up > down + EPS) {
	mid = (up+down)*0.5;
	v[0] = (1. - mid) * v0[0] + mid * v1[0];
	v[1] = (1. - mid) * v0[1] + mid * v1[1];
	v[2] = (1. - mid) * v0[2] + mid * v1[2];
	if ((d = dist(v, center)) < r - EPS)
	    down = mid;	/* mid is inside */
	else if (d > r + EPS)
	    up = mid;	/* mid is outside */
	else
	    break;	/* mid is on the surface */
    }
    mid = (down + up) * 0.5;
    if (mid < tol || mid > 1.0 - tol)
	return -1.0;
    return swap ? 1.0 - mid : mid;
#else
    /* compute cuts analytically by solving the quadratic equation:
     *	(v0-center,v0-center) + 2(v1-v0,v0-center)x + (v1-v0,v1-v0)x^2 = r^2 */
    FLOAT a, b, c, d, x1, x2;

    if (center == NULL) {
	v[0] = v[1] = v[2] = 0.0;
	center = v;
    }

    c = dist2(v0, center) - r * r;
    b = ((v1[0]-v0[0]) * (v0[0] - center[0]) +
	 (v1[1]-v0[1]) * (v0[1] - center[1]) +
	 (v1[2]-v0[2]) * (v0[2] - center[2])) * 2.0;
    a = dist2(v1, v0);
    /* solve the equation: a x^2 + b x + c = 0 */
    d = b * b - 4. * a * c;	/* d := b^2 - 4ac */
    if (d <= 0.0)
	return -1.0;	/* no root or one double root */
    d = Sqrt(d);
    x1 = (-b + d) / (a + a);
    x2 = (-b - d) / (a + a);
    /* Note: return a cut when exactly one of x1 and x2 is in [0+tol, 1-tol]
     * and the other is not in the range [-tol, 1+tol] */
    if (x1 >= tol && x1 <= 1. - tol)
	return (x2 >= -tol && x2 <= 1. + tol) ? -1.0 : x1;
    else if (x2 >= tol && x2 <= 1. - tol)
	return (x1 >= -tol && x1 <= 1. + tol) ? -1.0 : x2;
    else
	return -1;
#endif
}

static INT
getSurfaceCuts(GRID *g, FLOAT *cuts, FLOAT *center, FLOAT r, FLOAT tol)
/* returns number of cuts */
{
    INT i, count;
    ELEMENT *e;
    int v0, v1;
    VEF_MAP *m = phgDofSetupVEFMap(g, NULL, EDGE_FLAG);

    assert(m != NULL || g->nedge == 0);
    for (count = i = 0; i < g->nedge; i++) {
	int edge = m->Eind[i];
	e = m->Emap[i];
	if (e == NULL) {
	    /* unreferenced edge no. */
	    cuts[i] = -1.0;
	    continue;
	}
	GetEdgeVertices(e, edge, v0, v1)
	cuts[i] = getCutPosition(g->verts[e->verts[v0]], g->verts[e->verts[v1]],
				     center, r, tol);
	if (cuts[i] > 0.0 && (g->types_edge[i] & OWNER))
	    count++;
    }
    phgDofFreeVEFMap(&m);

    return count;
}

int
main(int argc, char **argv)
/* FIXME (surface-cut bug?): cut_test1 -tol 1e-2 cut_test1.dat */
{
    GRID *g, *g_new;
    ELEMENT *e;
    FLOAT r = 1.0, tol = 1e-6, *cuts, vol;
    INT loop = 1, mem_max = 1024, count;
    BOOLEAN vtk = FALSE;
    size_t mem;

    phgOptionsRegisterInt("-loop", "Loop count", &loop);
    phgOptionsRegisterInt("-mem_max", "Maximum memory (MB)", &mem_max);
    phgOptionsRegisterFloat("-r", "Radius of the sphere", &r);
    phgOptionsRegisterFloat("-tol", "Tolerance for computing the cuts", &tol);
    phgOptionsRegisterNoArg("-vtk", "Whether create VTK file", &vtk);

    phgInit(&argc, &argv);

    if (argc != 2) {
	phgPrintf("Usage: %s mesh_file\n", argv[0]);
	phgFinalize();
	exit(1);
    }
    
    g = phgNewGrid(-1);
    if (!phgImport(g, argv[1], TRUE))
	phgError(1, "can't read file \"%s\".\n", argv[1]);

    for (mem = 0; loop > 0 && mem < mem_max * (size_t)1024 * 1024; loop--) {
	cuts = phgAlloc(sizeof(*cuts) * g->nedge);
	count = getSurfaceCuts(g, cuts, NULL, r, tol);
	g_new = phgSurfaceCut(g, cuts);
	phgFree(cuts);
	/* Mark elements */
	vol = 0.0;
	ForAllElements(g_new, e) {
	    FLOAT vc[3], lambda[] = {0.25, 0.25, 0.25, 0.25};
	    phgGeomLambda2XYZ(g_new, e, lambda, vc, vc + 1, vc + 2);
	    if (dist2(vc, NULL) < r * r) {
		e->region_mark = 1;
		vol += phgGeomGetVolume(g_new, e);
	    }
	    else {
		e->region_mark = 0;
	    }
	}
#if USE_MPI
	if (g_new->nprocs > 1) {
	    FLOAT d;
	    INT i;
	    d = vol;
	    MPI_Reduce(&d, &vol, 1, PHG_MPI_FLOAT, MPI_SUM, 0, g_new->comm);
	    i = count;
	    MPI_Reduce(&i, &count, 1, PHG_MPI_INT, MPI_SUM, 0, g_new->comm);
	}
#endif	/* USE_MPI */
	phgMemoryUsage(g_new, &mem);
	if (phgRank == 0) {
	    printf("loop = %d:\n", loop);
	    printf("    Number of cut points = %d\n", count);
	    printf("    Number of elements = %d\n", g_new->nelem_global);
	    printf("    Number of vertices = %d\n", g_new->nvert_global);
	    printf("    Number of edges = %d\n", g_new->nedge_global);
	    printf("    Number of faces = %d\n", g_new->nface_global);
	    printf("    Memory used = %dMB\n",
					(int)((mem+1024*512)/(1024*1024)));
	    printf("    Volume = %g\n", (double)vol);
	}
	if (loop > 1) {
	    /*phgRefineRandomElements(g, "50%");*/
	    phgRefineAllElements(g, 1);
	    phgBalanceGrid(g, 1.0, 0, NULL, 0.0);
	}
	else if (vtk) {
	    phgPrintf("\"%s\" created.\n",
			phgExportVTK(g_new, "cut_test1.vtk", NULL));
	}
	phgFreeGrid(&g_new);
    }

    phgFreeGrid(&g);
    phgFinalize();

    return 0;
}
